/**********************************************************************
 *
 * papi_cmd.c:  PAPI command line interface /test program
 *
 * Copyright (c) 2002 International Business Machines Corp.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include "papi.h"

static char             attrs_buff[4096];

//*********************************************************************
//  syntax
//*********************************************************************
void syntax()
{
	printf("Open Standard Print API (PAPI) command line\n\n");
	printf("Syntax:  papi -p prtr filename [\"job-attrs\"|-X attr-file|-J job-ticket-file]\n");
	printf("    or:  papi -ps prtr [\"job-attrs\"]\n");
	printf("    or:  papi -v prtr filename [\"job-attrs\"|-X attr-file|-J job-ticket-file]\n");
    printf("    or:  papi -l [filter-value]\n");
	printf("    or:  papi -lj prtr [req-attr1 [req-attr2] ...] \n");
	printf("    or:  papi -lJ prtr [req-attr1 [req-attr2] ...] \n");
    printf("    or:  papi -dj prtr\n");
    printf("    or:  papi -q prtr [req-attr1 [req-attr2] ...] \n");
    printf("    or:  papi -m prtr \"printer-attrs\" |-X attr-file\n");
    printf("    or:  papi -qj prtr job-id [req-attr1 [req-attr2] ...] \n");
    printf("    or:  papi -cj prtr job-id \n");
    printf("    or:  papi -mj prtr job-id \"job-attrs\"|-X attr-file\n");
    printf("    or:  papi -hj prtr job-id [hold-until-string]\n");
    printf("    or:  papi -rj prtr job-id \n");
    printf("    or:  papi -Rj prtr job-id \n");
	printf("    or:  papi -P prtr\n");
	printf("    or:  papi -R prtr\n");
	printf("Where:\n");
	printf("   -p        = submit job to specified printer\n");
	printf("   -ps       = submit job from stdin to specified printer\n");
	printf("   -v        = validate job to specified printer\n");
	printf("   -l        = list printers\n");
	printf("   -lj       = list jobs for printer (not completed)\n");
	printf("   -lJ       = list jobs for printer (completed)\n");
	printf("   -dj       = delete all jobs (purge) for printer\n");
	printf("   -q        = query attributes of specified printer\n");
	printf("   -m        = modify attributes of specified printer\n");
	printf("   -qj       = query attributes of specified job\n");
	printf("   -cj       = cancel the specified job\n");
	printf("   -mj       = modify the specified job\n");
	printf("   -hj       = hold the specified job\n");
	printf("   -rj       = release the specified job\n");
	printf("   -Rj       = restart the specified job\n");
	printf("   -P        = pause the specified printer\n");
	printf("   -R        = resume the specified printer\n");
	printf("   req-attrN = space delimited requested attributes (query all if none specified)\n");
	printf("\nNOTE: Environment variable IPPSERVER can be set to the IPP\n");
	printf("server to send all requests to.  If this is not set, the default\n");
	printf("of ipp://localhost:631 is used.\n");
	exit(1);
}

typedef enum
{
	LIST_PRINTERS,
	SUBMIT_JOB,
	SUBMIT_JOB_STDIN,
	VALIDATE_JOB,
	QUERY_PRINTER,
	MODIFY_PRINTER,
	QUERY_JOB,
	CANCEL_JOB,
	MODIFY_JOB,
	LIST_JOBS,
	LIST_JOBS_NOT_COMPLETED,
	PURGE_JOBS,
	PAUSE_PRINTER,
	RESUME_PRINTER,
	HOLD_JOB,
	RELEASE_JOB,
	RESTART_JOB,
	DEBUG_TEST,
} oper_type_t;

//*********************************************************************
//  listAllAttributes
//*********************************************************************
void listAllAttributes(papi_attribute_t** attrs)
{
	papi_status_t rc;
	char buffer[8192];

	rc = papiAttributeListToString(attrs, "\n", buffer, 8192);
	if (rc == PAPI_OK)
	{
		printf("%s\n", buffer);
	}
	else
	{
		fprintf(stderr, "papiAttributeListToString failed (rc=%d): %s\n", rc, papiStatusString(rc));
	}
}

//*********************************************************************
//  listSomeAttributes
//*********************************************************************
void listSomeAttributes(papi_attribute_t** attrs, const char* req_attrs[])
{
	papi_attribute_t*	attr_list[500];
	papi_attribute_t*	tmp;

	int i;
	int k;
	if (attrs == NULL)
	{
		printf("No attributes in response\n");
		return;
	}

	/* Find the first of each */
	for(k=0, i=0; req_attrs[i] != NULL; i++)
	{
	    tmp = papiAttributeListFind(attrs, req_attrs[i]);
		if (tmp)
			attr_list[k++] = tmp;
	}
	attr_list[k] = NULL;

	/* Now go through and list 'em all */
	listAllAttributes(attr_list);
}

//*********************************************************************
//  getAttrsString
//*********************************************************************
char* getAttrsString(int arg_i, int argc, char** argv)
{
	FILE*            attrsf;
	int				 byte_count;
	char*            p_attrs = argv[arg_i];

	if ((argc == (arg_i+2)) && (p_attrs != NULL) &&
	    (p_attrs[0] == '-'))
	{
		if (p_attrs[1] == 'X')
		{
			attrsf = fopen(argv[arg_i+1], "r");	
			if (attrsf == NULL)
			{
				printf("Cannot open attributes file %s\n", argv[5]);
				exit(1);
			}
			byte_count = fread(attrs_buff, 1, 4095, attrsf);
			attrs_buff[byte_count] = '\0';
			p_attrs = attrs_buff;
			fclose(attrsf);
		}
		else if (p_attrs[1] == 'J')
		{
			p_attrs = NULL;
		}
	}

	return(p_attrs);
}

//*********************************************************************
//  checkResponse
//*********************************************************************
int checkResponse(papi_status_t rc, papi_service_t handle, const char* req_str)
{
	int res = 1;
	if (rc != PAPI_OK)
	{
		fprintf(stderr, "%s failed (rc=%d): %s\n", req_str, rc, papiStatusString(rc));
		if (handle != NULL)
			fprintf(stderr, "       details: %s\n", papiServiceGetStatusMessage(handle));
		res = 0;
	}

	return(res);
}

//*********************************************************************
//                           main
//*********************************************************************
int main(int argc, char *argv[])
{
	papi_service_t handle;
	oper_type_t oper_type;
	papi_status_t rc;
	int i;
	char* req_attrs[200];
	char* ipp_server = "ipp://localhost:631";
	char* p_env;
	char server_uri[HTTP_MAX_URI];
	char method[HTTP_MAX_URI];
	char username[HTTP_MAX_URI];
	char host[HTTP_MAX_URI];
	char resource[HTTP_MAX_URI];
	int port;
	papi_attribute_t* attr;
	papi_attribute_t** attrs;
	papi_printer_t printer;
	papi_filter_t filter;
	static char* list_req_attrs[] =
	{
		"printer-name",
		NULL
	};
	static char* lj_req_attrs[] =
	{
		"job-name",
		NULL
	};
	char* doc_names[2];
	papi_printer_t*  printers;
	papi_job_t       job;
	papi_job_t*      jobs;
	char*            tmp_str;
	char*            p_attrs;
	papi_job_ticket_t* job_ticket = NULL;

	papi_stream_t	 stream;
	char			 buffer[4096];
	int				 byte_count;

	//-----------------------------------------------------------------
	// Get or default the server information
	//-----------------------------------------------------------------
	p_env = getenv("IPPSERVER");
	if (p_env)
		ipp_server = p_env;
	httpSeparate(server_uri, method, username, host, &port, resource);

	//-----------------------------------------------------------------
	// Validate arguments and set the test type
	//-----------------------------------------------------------------
	if (argc == 1)
		syntax();
	if (!strcmp(argv[1], "-p"))
		oper_type = SUBMIT_JOB;
	else if (!strcmp(argv[1], "-ps"))
		oper_type = SUBMIT_JOB_STDIN;
	else if (!strcmp(argv[1], "-v"))
		oper_type = VALIDATE_JOB;
	else if (!strcmp(argv[1], "-l"))
		oper_type = LIST_PRINTERS;
	else if (!strcmp(argv[1], "-lj"))
		oper_type = LIST_JOBS_NOT_COMPLETED;
	else if (!strcmp(argv[1], "-lJ"))
		oper_type = LIST_JOBS;
	else if (!strcmp(argv[1], "-q"))
		oper_type = QUERY_PRINTER;
	else if (!strcmp(argv[1], "-DD"))
		oper_type = DEBUG_TEST;
	else if (!strcmp(argv[1], "-m"))
		oper_type = MODIFY_PRINTER;
	else if (!strcmp(argv[1], "-qj"))
		oper_type = QUERY_JOB;
	else if (!strcmp(argv[1], "-cj"))
		oper_type = CANCEL_JOB;
	else if (!strcmp(argv[1], "-mj"))
		oper_type = MODIFY_JOB;
	else if (!strcmp(argv[1], "-hj"))
		oper_type = HOLD_JOB;
	else if (!strcmp(argv[1], "-rj"))
		oper_type = RELEASE_JOB;
	else if (!strcmp(argv[1], "-Rj"))
		oper_type = RESTART_JOB;
	else if (!strcmp(argv[1], "-dj"))
		oper_type = PURGE_JOBS;
	else if (!strcmp(argv[1], "-P"))
		oper_type = PAUSE_PRINTER;
	else if (!strcmp(argv[1], "-R"))
		oper_type = RESUME_PRINTER;
	else
		syntax();

	if ((oper_type == QUERY_PRINTER) && (argc < 3))
		syntax();
	if ((oper_type == SUBMIT_JOB) && (argc < 4))
		syntax();
	if ((oper_type == SUBMIT_JOB_STDIN) && (argc < 3))
		syntax();
	if ((oper_type == MODIFY_PRINTER) && (argc < 4))
		syntax();
	if (((oper_type == LIST_JOBS)||(oper_type == LIST_JOBS_NOT_COMPLETED)) && (argc < 3))
		syntax();
	if ((oper_type == LIST_PRINTERS) && (argc > 3))
		syntax();
	if ((oper_type == QUERY_JOB) && (argc < 4))
		syntax();
	if ((oper_type == CANCEL_JOB) && (argc < 3))
		syntax();
	if ((oper_type == MODIFY_JOB) && (argc < 4))
		syntax();
	if ((oper_type == HOLD_JOB) && (argc < 3))
		syntax();
	if ((oper_type == RELEASE_JOB) && (argc < 3))
		syntax();
	if ((oper_type == RESTART_JOB) && (argc < 3))
		syntax();
	if ((oper_type == PURGE_JOBS) && (argc < 3))
		syntax();
	if ((oper_type == VALIDATE_JOB) && (argc < 4))
		syntax();
	if ((oper_type == PAUSE_PRINTER) && (argc < 3))
		syntax();
	if ((oper_type == RESUME_PRINTER) && (argc < 3))
		syntax();

	//-----------------------------------------------------------------
	// Common init
	//-----------------------------------------------------------------
	rc = papiServiceCreate(&handle, NULL, NULL, NULL, NULL, PAPI_ENCRYPT_NEVER, NULL);
	if (!checkResponse(rc, handle, "papiServiceCreate"))
	{
		if (handle != NULL)
			papiServiceDestroy(handle);
		return(__LINE__);
	}
	if (rc != PAPI_OK)
	{
		fprintf(stderr, "papiServiceCreate failed: %s\n", papiStatusString(rc));
		if (handle != NULL)
		{
			fprintf(stderr, "       details: %s\n", papiServiceGetStatusMessage(handle));
			papiServiceDestroy(handle);
		}
		return(1);
	}

	// Perform the selected test
	switch(oper_type)
	{
	//-----------------------------------------------------------------
	// Submit job
	//-----------------------------------------------------------------
	case SUBMIT_JOB:
		doc_names[0] = argv[3];
		doc_names[1] = NULL;
		
		/* Parse input string into PAPI attrs */
		attrs = NULL;
		p_attrs = getAttrsString(4, argc, argv);
		if ( papiAttributeListFromString(&attrs, PAPI_ATTR_APPEND, p_attrs)
		     != PAPI_OK )
		{
			printf("Error parsing input attributes\n");
			return(1);
		}

		/* Check for a job ticket */
		if ((argc == 6) && (argv[4] != NULL) &&
	        (argv[4][0] == '-') && (argv[4][1] == 'J'))
		{
			job_ticket = malloc(sizeof(papi_job_ticket_t));
			job_ticket->format = PAPI_JT_FORMAT_JDF;
			job_ticket->ticket_data = NULL;
			job_ticket->file_name = argv[5];
		}

		/* Submit the job */
	 	rc = papiJobSubmit(handle, argv[2], attrs, job_ticket, doc_names, &job);
        papiAttributeListFree(attrs);
        if (job_ticket)
        	free(job_ticket);
		if (!checkResponse(rc, handle, "papiJobSubmit"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}

		printf("Job submitted successfully to %s: ", argv[2]);
		if (job)
		{
			printf(" job-id=%d\n", papiJobGetId(job));
			papiJobFree(job);
		}
		else
		{
			printf("<no job returned>\n");
		}
		
		break;

	//-----------------------------------------------------------------
	// Submit job from stdin
	//-----------------------------------------------------------------
	case SUBMIT_JOB_STDIN:
		
		/* Parse input string into PAPI attrs */
		attrs = NULL;
		p_attrs = getAttrsString(3, argc, argv);
		if ( papiAttributeListFromString(&attrs, PAPI_ATTR_APPEND, p_attrs)
		     != PAPI_OK )
		{
			printf("Error parsing input attributes\n");
			return(1);
		}

		/* Check for a job ticket */
		if ((argc == 5) && (argv[3] != NULL) &&
	        (argv[3][0] == '-') && (argv[3][1] == 'J'))
		{
			job_ticket = malloc(sizeof(papi_job_ticket_t));
			job_ticket->format = PAPI_JT_FORMAT_JDF;
			job_ticket->ticket_data = NULL;
			job_ticket->file_name = argv[4];
		}

		/* Open the job stream */
	 	rc = papiJobStreamOpen(handle, argv[2], attrs, job_ticket, &stream);
        papiAttributeListFree(attrs);
        if (job_ticket)
        	free(job_ticket);
		if (!checkResponse(rc, handle, "papiJobStreamOpen"))
		{
			papiJobStreamClose(handle, stream, &job);
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}

		/* Read/write stdin to the job stream */
		byte_count = fread(buffer, 1, 4096, stdin);
		while(byte_count > 0)
		{
		 	rc = papiJobStreamWrite(handle, stream, buffer, byte_count);
			if (!checkResponse(rc, handle, "papiJobStreamWrite"))
			{
				papiJobStreamClose(handle, stream, &job);
				papiJobFree(job);
				papiServiceDestroy(handle);
				return(__LINE__);
			}

			byte_count = fread(buffer, 1, 4096, stdin);
		}

		/* Close/submit the job */
	 	rc = papiJobStreamClose(handle, stream, &job);
        papiAttributeListFree(attrs);
		if (!checkResponse(rc, handle, "papiJobStreamClose"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}

		printf("Job submitted successfully to %s: ", argv[2]);
		if (job)
		{
			printf(" job-id=%d\n", papiJobGetId(job));
			papiJobFree(job);
		}
		else
		{
			printf("<no job returned>\n");
		}
		
		break;

	//-----------------------------------------------------------------
	// Validate job
	//-----------------------------------------------------------------
	case VALIDATE_JOB:
		doc_names[0] = argv[3];
		doc_names[1] = NULL;
		
		/* Parse input string into PAPI attrs */
		attrs = NULL;
		p_attrs = getAttrsString(4, argc, argv);
		if ( papiAttributeListFromString(&attrs, PAPI_ATTR_APPEND, p_attrs)
		     != PAPI_OK )
		{
			printf("Error parsing input attributes\n");
			return(1);
		}

		/* Check for a job ticket */
		if ((argc == 6) && (argv[4] != NULL) &&
	        (argv[4][0] == '-') && (argv[4][1] == 'J'))
		{
			job_ticket = malloc(sizeof(papi_job_ticket_t));
			job_ticket->format = PAPI_JT_FORMAT_JDF;
			job_ticket->ticket_data = NULL;
			job_ticket->file_name = argv[5];
		}

		/* Validate the job */
	 	rc = papiJobValidate(handle, argv[2], attrs, job_ticket, doc_names, &job);
        papiAttributeListFree(attrs);
        if (job_ticket)
        	free(job_ticket);
		if (!checkResponse(rc, handle, "papiJobValidate"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}

		printf("Job validated successfully to %s\n", argv[2]);
		papiJobFree(job);
		
		break;

	//-----------------------------------------------------------------
	// List printers
	//-----------------------------------------------------------------
	case LIST_PRINTERS:
		if (argc > 2)
		{
			filter.type = PAPI_FILTER_BITMASK;	
			filter.filter.bitmask.mask = atoi(argv[2]);
			filter.filter.bitmask.value = filter.filter.bitmask.mask;
		 	rc = papiPrintersList(handle, list_req_attrs, &filter, &printers);
		}
		else
		{
		 	rc = papiPrintersList(handle, list_req_attrs, NULL, &printers);
		}
	
		if (!checkResponse(rc, handle, "papiPrintersList"))
		{
			papiPrinterListFree(printers);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
		
		printf("Printers:\n");
		for(i=0; printers[i] != NULL; i++)
		{
			attr = papiAttributeListFind( papiPrinterGetAttributeList(printers[i]),
							list_req_attrs[0]);
			if (attr)
				printf("   %s\n", attr->values[0]->string);
		}

		papiPrinterListFree(printers);

		break;

	//-----------------------------------------------------------------
	// Query printer
	//-----------------------------------------------------------------
	case QUERY_PRINTER:
		if (argc > 3)
		{
			for(i=3; i<argc; i++)
				req_attrs[i-3] = argv[i];
			req_attrs[i-3] = NULL;
		}
		else
		{
			req_attrs[0] = NULL;
		}

	 	rc = papiPrinterQuery(handle, argv[2], req_attrs, NULL, &printer);

		if (!checkResponse(rc, handle, "papiPrinterQuery"))
		{
			papiPrinterFree(printer);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("%s attributes:\n", argv[2]);
		if (req_attrs[0] != NULL)
			listSomeAttributes(papiPrinterGetAttributeList(printer), req_attrs);
		else
			listAllAttributes(papiPrinterGetAttributeList(printer));

		papiPrinterFree(printer);
		break;

	//-----------------------------------------------------------------
	// Modify printer
	//-----------------------------------------------------------------
	case MODIFY_PRINTER:
		/* Parse input string into PAPI attrs */
		attrs = NULL;
		p_attrs = getAttrsString(3, argc, argv);
		if ( papiAttributeListFromString(&attrs, PAPI_ATTR_APPEND, p_attrs)
		     != PAPI_OK )
		{
			printf("Error parsing input attributes\n");
			return(1);
		}

		/* Modify the printer */
	 	rc = papiPrinterModify(handle, argv[2], attrs, &printer);

		if (!checkResponse(rc, handle, "papiPrinterModify"))
		{
			papiPrinterFree(printer);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
		
		printf("Printer %s modified successfully\n", argv[2]);
	
		papiPrinterFree(printer);
		break;

	//-----------------------------------------------------------------
	// Pause printer
	//-----------------------------------------------------------------
	case PAUSE_PRINTER:

	 	rc = papiPrinterPause(handle, argv[2]);

		if (!checkResponse(rc, handle, "papiPrinterPause"))
		{
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("%s paused\n", argv[2]);

		break;

	//-----------------------------------------------------------------
	// Resume printer
	//-----------------------------------------------------------------
	case RESUME_PRINTER:

	 	rc = papiPrinterResume(handle, argv[2]);

		if (!checkResponse(rc, handle, "papiPrinterResume"))
		{
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("%s resumed\n", argv[2]);
	
		break;

	//-----------------------------------------------------------------
	// Purge jobs
	//-----------------------------------------------------------------
	case PURGE_JOBS:

	 	rc = papiPrinterPurgeJobs(handle, argv[2], NULL);

		if (!checkResponse(rc, handle, "papiPrinterPurgeJobs"))
		{
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("All jobs purged from printer %s\n", argv[2]);
	
		break;

	//-----------------------------------------------------------------
	// List jobs
	//-----------------------------------------------------------------
	case LIST_JOBS:
	case LIST_JOBS_NOT_COMPLETED:
		if (argc > 3)
		{
			for(i=3; i<argc; i++)
				req_attrs[i-3] = argv[i];
			req_attrs[i-3] = NULL;
		}
		else
		{
			req_attrs[0] = NULL;
		}
	
		if (oper_type == LIST_JOBS)
		 	rc = papiPrinterListJobs(handle,
							argv[2],
							req_attrs,
							PAPI_LIST_JOBS_OTHERS|PAPI_LIST_JOBS_COMPLETED,
							0,
							&jobs);
		else
		 	rc = papiPrinterListJobs(handle,
							argv[2],
							req_attrs,
							PAPI_LIST_JOBS_OTHERS|PAPI_LIST_JOBS_NOT_COMPLETED,
							0,
							&jobs);

		if (!checkResponse(rc, handle, "papiPrinterListJobs"))
		{
			papiJobListFree(jobs);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
		
		for(i=0; jobs[i] != NULL; i++)
		{
			printf("Job:\n");
			if (req_attrs[0] != NULL)
				listSomeAttributes(papiJobGetAttributeList(jobs[i]), req_attrs);
			else
				listAllAttributes(papiJobGetAttributeList(jobs[i]));
		}
		if (i == 0)
			printf("no jobs found\n");

		papiJobListFree(jobs);
	
		break;

	//-----------------------------------------------------------------
	// Query job
	//-----------------------------------------------------------------
	case QUERY_JOB:
		if (argc > 4)
		{
			for(i=4; i<argc; i++)
				req_attrs[i-4] = argv[i];
			req_attrs[i-4] = NULL;
		}
		else
		{
			req_attrs[0] = NULL;
		}

	 	rc = papiJobQuery(handle, argv[2], atoi(argv[3]), req_attrs, &job);

		if (!checkResponse(rc, handle, "papiJobQuery"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s attributes:\n", argv[2]);
		if (req_attrs[0] != NULL)
			listSomeAttributes(papiJobGetAttributeList(job), req_attrs);
		else
			listAllAttributes(papiJobGetAttributeList(job));

		papiJobFree(job);
		break;

	//-----------------------------------------------------------------
	// Cancel job
	//-----------------------------------------------------------------
	case CANCEL_JOB:

	 	rc = papiJobCancel(handle, argv[2], atoi(argv[3]));

		if (!checkResponse(rc, handle, "papiJobCancel"))
		{
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s successfully cancelled\n", argv[3]);
		break;

	//-----------------------------------------------------------------
	// Modify job
	//-----------------------------------------------------------------
	case MODIFY_JOB:
		/* Parse input string into PAPI attrs */
		attrs = NULL;
		p_attrs = getAttrsString(4, argc, argv);
		if ( papiAttributeListFromString(&attrs, PAPI_ATTR_APPEND, p_attrs)
		     != PAPI_OK )
		{
			printf("Error parsing input attributes\n");
			return(1);
		}

		/* Modify the job */
	 	rc = papiJobModify(handle, argv[2], atoi(argv[3]), attrs, &job);

		if (!checkResponse(rc, handle, "papiJobModify"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s successfully modified\n", argv[3]);
		papiJobFree(job);
		break;

	//-----------------------------------------------------------------
	// Hold job
	//-----------------------------------------------------------------
	case HOLD_JOB:

		if (argc > 4)
		 	rc = papiJobHold(handle, argv[2], atoi(argv[3]), argv[4], NULL);
		else
		 	rc = papiJobHold(handle, argv[2], atoi(argv[3]), NULL, NULL);

		if (!checkResponse(rc, handle, "papiJobHold"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s successfully held\n", argv[3]);
		break;

	//-----------------------------------------------------------------
	// Release job
	//-----------------------------------------------------------------
	case RELEASE_JOB:

	 	rc = papiJobRelease(handle, argv[2], atoi(argv[3]));

		if (!checkResponse(rc, handle, "papiJobRelease"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s successfully released\n", argv[3]);
		break;

	//-----------------------------------------------------------------
	// Restart job
	//-----------------------------------------------------------------
	case RESTART_JOB:

	 	rc = papiJobRestart(handle, argv[2], atoi(argv[3]));

		if (!checkResponse(rc, handle, "papiJobRestart"))
		{
			papiJobFree(job);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		printf("Job %s successfully restarted\n", argv[3]);
		break;

	//-----------------------------------------------------------------
	// Debug test: use this (-DD) to test whatever misc PAPI code
	// needs to be tested by replacing this case
	//-----------------------------------------------------------------
	case DEBUG_TEST:
		/* This tests the papiAttributeListGet* functions */
	 	rc = papiPrinterQuery(handle, argv[2], NULL, NULL, &printer);

		if (!checkResponse(rc, handle, "papiPrinterQuery"))
		{
			papiPrinterFree(printer);
			papiServiceDestroy(handle);
			return(__LINE__);
		}
	
		attrs = papiPrinterGetAttributeList(printer);
		if (attrs != NULL)
		{
			rc = papiAttributeListGetString(attrs, NULL, "printer-name", &tmp_str);
			if (!checkResponse(rc, handle, "papiAttributeListGetString"))
			{
				papiPrinterFree(printer);
				papiServiceDestroy(handle);
				return(__LINE__);
			}
			printf("printer-name: %s\n", tmp_str);

			rc = papiAttributeListGetInteger(attrs, NULL, "copies-default", &i);
			if (!checkResponse(rc, handle, "papiAttributeListGetInteger"))
			{
				papiPrinterFree(printer);
				papiServiceDestroy(handle);
				return(__LINE__);
			}
			printf("copies-default: %d\n", i);

			rc = papiAttributeListGetRange(attrs, NULL, "printer-type", &i, &i);
			if (rc == PAPI_OK)
			{
				printf("Expected error on printer-type returned success!\n");
				papiPrinterFree(printer);
				papiServiceDestroy(handle);
				return(__LINE__);
			}
		}
		else
		{
			printf("no attributes returned!\n");
		}

		papiPrinterFree(printer);
		break;

	default:
		syntax();
	}

	//-----------------------------------------------------------------
	// Common clean up
	//-----------------------------------------------------------------
	papiServiceDestroy(handle);
		
	return(0);
}
